@@ -30,13 +30,14 @@
 
 use super::super::utils;
 use super::{BaseOptimizer, BuildOptimizer, OptimizerStepError, TrustRegionOptimizer};
-use crate::logging::{Logger, LoggerHelper};
+use crate::logging::StatsLogger;
+use serde::{Deserialize, Serialize};
 use std::borrow::Borrow;
-use std::convert::{Infallible, TryInto};
-use tch::{nn::VarStore, Tensor};
+use std::convert::Infallible;
+use tch::Tensor;
 
 /// Configuration for the Conjugate Gradient Optimizer
-#[derive(Debug, Clone, Copy, PartialEq)]
+#[derive(Debug, Clone, Copy, PartialEq, Serialize, Deserialize)]
 pub struct ConjugateGradientOptimizerConfig {
     /// Number of CG iterations used to calculate A^-1 g
     pub iterations: u64,
@@ -67,8 +68,14 @@ impl BuildOptimizer for ConjugateGradientOptimizerConfig {
     type Optimizer = ConjugateGradientOptimizer;
     type Error = Infallible;
 
-    fn build_optimizer(&self, vs: &VarStore) -> Result<Self::Optimizer, Self::Error> {
-        Ok(ConjugateGradientOptimizer::new(vs, *self))
+    fn build_optimizer<'a, I>(&self, variables: I) -> Result<Self::Optimizer, Self::Error>
+    where
+        I: IntoIterator<Item = &'a Tensor>,
+    {
+        Ok(ConjugateGradientOptimizer::new(
+            variables.into_iter().map(Tensor::shallow_clone).collect(),
+            *self,
+        ))
     }
 }
 
@@ -87,9 +94,9 @@ pub struct ConjugateGradientOptimizer {
 }
 
 impl ConjugateGradientOptimizer {
-    pub fn new(vs: &VarStore, config: ConjugateGradientOptimizerConfig) -> Self {
+    pub fn new(variables: Vec<Tensor>, config: ConjugateGradientOptimizerConfig) -> Self {
         Self {
-            params: vs.trainable_variables(),
+            params: variables,
             config,
         }
     }
@@ -108,7 +115,7 @@ impl TrustRegionOptimizer for ConjugateGradientOptimizer {
         &self,
         loss_distance_fn: &dyn Fn() -> (Tensor, Tensor),
         max_distance: f64,
-        logger: &mut dyn Logger,
+        logger: &mut dyn StatsLogger,
     ) -> Result<f64, OptimizerStepError> {
         let (loss, distance) = loss_distance_fn();
 
@@ -150,7 +157,7 @@ impl TrustRegionOptimizer for ConjugateGradientOptimizer {
             x if x.is_nan() => 1.0,
             x => x,
         };
-        logger.unwrap_log("step_size", step_size);
+        logger.log_scalar("step_size", step_size);
 
         let descent_step = step_size * step_dir;
         let initial_loss: f64 = loss.into();
@@ -176,7 +183,7 @@ impl ConjugateGradientOptimizer {
         loss_constraint_fn: &F,
         max_constraint_value: f64,
         initial_loss: f64,
-        logger: &mut dyn Logger,
+        logger: &mut dyn StatsLogger,
     ) -> Result<(), OptimizerStepError>
     where
         F: Fn() -> (Tensor, Tensor) + ?Sized,
@@ -189,7 +196,7 @@ impl ConjugateGradientOptimizer {
 
         let mut loss = initial_loss;
         let mut constraint_val = f64::INFINITY;
-        logger.unwrap_log("initial_loss", loss);
+        logger.log_scalar("loss_initial", loss);
         for i in 0..self.config.max_backtracks {
             let ratio = self.config.backtrack_ratio.powi(i.try_into().unwrap());
 
@@ -208,14 +215,14 @@ impl ConjugateGradientOptimizer {
             loss = loss_tensor.into();
             constraint_val = constraint_tensor.into();
             if loss < initial_loss && constraint_val <= max_constraint_value {
-                logger.unwrap_log("num_backtracks", i as f64);
-                logger.unwrap_log("step_scale", ratio);
+                logger.log_scalar("num_backtracks", i as f64);
+                logger.log_scalar("step_scale", ratio);
                 break;
             }
         }
 
-        logger.unwrap_log("final_loss", loss);
-        logger.unwrap_log("final_constraint_val", constraint_val);
+        logger.log_scalar("loss_final", loss);
+        logger.log_scalar("constraint_val_final", constraint_val);
 
         let result = if loss.is_nan() {
             Err(OptimizerStepError::NaNLoss)
@@ -359,8 +366,9 @@ impl MatrixVectorProduct for Tensor {
 /// # Reference
 /// * <https://en.wikipedia.org/wiki/Conjugate_gradient_method>
 /// * <https://github.com/rlworkgroup/garage/blob/90b60905b29cea8f8373c6732ced0cadf8489b0c/src/garage/torch/optimizers/conjugate_gradient_optimizer.py>
+#[allow(non_snake_case)]
 fn solve_conjugate_gradient<T: MatrixVectorProduct<Vector = Tensor>>(
-    #[allow(non_snake_case)] f_Ax: &T,
+    f_Ax: &T,
     b: &Tensor,
     iterations: u64,
     residual_tol: f64,
@@ -433,9 +441,8 @@ mod cg_optimizer {
     fn shared_loss_distance_computation() {
         let config = ConjugateGradientOptimizerConfig::default();
 
-        let vs = VarStore::new(Device::Cpu);
-        let x = vs.root().ones("x", &[2]);
-        let optimizer = config.build_optimizer(&vs).unwrap();
+        let x = Tensor::ones(&[2], (Kind::Float, Device::Cpu)).requires_grad_(true);
+        let optimizer = config.build_optimizer([&x]).unwrap();
 
         let y_prev = x.square().mean(Kind::Float).detach();
         let loss_distance_fn = || {
@@ -468,10 +475,9 @@ mod cg_optimizer {
     fn unused_params() {
         let config = ConjugateGradientOptimizerConfig::default();
 
-        let vs = VarStore::new(Device::Cpu);
-        let x = vs.root().ones("x", &[2]);
-        let _ = vs.root().zeros("unused", &[3]);
-        let optimizer = config.build_optimizer(&vs).unwrap();
+        let x = Tensor::ones(&[2], (Kind::Float, Device::Cpu)).requires_grad_(true);
+        let unused = Tensor::zeros(&[3], (Kind::Float, Device::Cpu)).requires_grad_(true);
+        let optimizer = config.build_optimizer([&x, &unused]).unwrap();
 
         let x_prev = x.detach().copy();
         let loss_distance_fn = || {
