@@ -7,8 +7,9 @@
 //!
 //! For example, the sequences `[0, 1, 2, 3]`, `[10, 11]`, `[100, 101]` are packed as
 //! `[0, 10, 100, 1, 11, 101, 2, 3]`.
+use super::sequence::Sequence;
 use std::iter;
-use std::ops::{Index, Range};
+use std::ops::Deref;
 use tch::{IndexOp, Scalar, Tensor};
 
 /// Iterator that packs together multiple sequences.
@@ -72,6 +73,10 @@ where
         U::Item: IntoIterator<IntoIter = T>,
     {
         let sequences: Vec<_> = sequences.into_iter().map(IntoIterator::into_iter).collect();
+        assert!(
+            sequences.windows(2).all(|w| w[0].len() >= w[1].len()),
+            "sequences not in monotonic decreasing order of length"
+        );
         let batch_size = sequences.len();
         Self {
             sequences,
@@ -105,190 +110,114 @@ where
     }
 
     fn size_hint(&self) -> (usize, Option<usize>) {
-        let len = self.sequences.iter().map(|s| s.len()).sum();
+        let len = self.sequences.iter().map(ExactSizeIterator::len).sum();
         (len, Some(len))
     }
 }
 
 impl<T> ExactSizeIterator for PackedIter<T> where T: ExactSizeIterator {}
 
-/// Iterator of indices that pack the specified contiguous sequences.
+/// Iterator that packs together the elements of multiple sequences.
+///
+/// Does not allocate any heap memory.
 ///
 /// # Example
 /// ```
-/// use relearn::utils::packed::PackingIndices;
+/// use relearn::utils::packed::PackedSeqIter;
 ///
-/// let data = [0, 1, 2, 3, 10, 11, 100, 101];
-/// let ranges = [0..4, 4..6, 6..8];
-/// let packed: Vec<_> = PackingIndices::from_sorted(&ranges).map(|i| data[i]).collect();
+/// let sequences: [&[_]; 3] = [&[0, 1, 2, 3], &[10, 11], &[100, 101]];
+/// let packed: Vec<_> = PackedSeqIter::from_sorted(&sequences).map(|&x| x).collect();
 /// assert_eq!(packed, vec![0, 10, 100, 1, 11, 101, 2, 3]);
 /// ```
 #[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
-pub struct PackingIndices<'a> {
-    /// Sequence index ranges, sorted in monotonically decreasing order of sequence length.
-    sequence_ranges: &'a [Range<usize>],
+pub struct PackedSeqIter<I> {
+    /// Initial copy of the sequence iterator. Never modified.
+    sequences: I,
 
     /// Current offset within the sequences.
     offset: usize,
-    /// Current sequence index within `sequences`.
-    seq_idx: usize,
-    /// Number of sequences with length longer than the current offset.
-    batch_size: usize,
+    /// Iterator of sequences for the current offset.
+    sequences_iter: I,
 }
 
-impl<'a> PackingIndices<'a> {
-    /// Create a new [`PackingIndices`] instance from ranges sorted in decreasing order of length.
-    ///
-    /// # Args
-    /// * `sequence_ranges` - Index range for each sequence,
-    ///     sorted in monotonically decreasing order of sequence length.
-    pub fn from_sorted(sequence_ranges: &'a [Range<usize>]) -> Self {
-        let mut batch_size = sequence_ranges.len();
-        let offset = 0;
-        update_batch_size(sequence_ranges, offset, &mut batch_size);
+impl<I> PackedSeqIter<I>
+where
+    I: Iterator + Clone,
+    <I as Iterator>::Item: Deref, // TODO: Try to avoid :Deref bound; use :Sequence directly
+    <<I as Iterator>::Item as Deref>::Target: Sequence,
+{
+    /// Initialize from sequences sorted in monotonic decreasing order of length.
+    pub fn from_sorted<T: IntoIterator<IntoIter = I>>(into_sequences: T) -> Self {
+        let sequences = into_sequences.into_iter();
+        assert!(
+            sequences
+                .clone()
+                .zip(sequences.clone().skip(1))
+                .all(|(a, b)| a.len() >= b.len()),
+            "sequences not in monotonic decreasing order of length"
+        );
+        let sequences_iter = sequences.clone();
         Self {
-            sequence_ranges,
-            offset,
-            seq_idx: 0,
-            batch_size,
+            sequences,
+            offset: 0,
+            sequences_iter,
         }
     }
 }
 
-impl<'a> Iterator for PackingIndices<'a> {
-    type Item = usize;
+impl<I> Iterator for PackedSeqIter<I>
+where
+    I: Iterator + ExactSizeIterator + Clone,
+    <I as Iterator>::Item: Deref,
+    <<I as Iterator>::Item as Deref>::Target: Sequence,
+{
+    type Item = <<I::Item as Deref>::Target as Sequence>::Item;
 
     fn next(&mut self) -> Option<Self::Item> {
-        if self.batch_size == 0 {
-            return None;
+        match self
+            .sequences_iter
+            .next()
+            .and_then(|seq| seq.get(self.offset))
+        {
+            Some(value) => Some(value), // Next value at this offset
+            None => {
+                // Increment offset and restart the loop
+                self.offset += 1;
+                self.sequences_iter = self.sequences.clone();
+                // If this fails then there are no more items left
+                self.sequences_iter
+                    .next()
+                    .and_then(|seq| seq.get(self.offset))
+            }
         }
-        let index = self.sequence_ranges[self.seq_idx].start + self.offset;
-        self.seq_idx += 1;
-        if self.seq_idx >= self.batch_size {
-            self.seq_idx = 0;
-            self.offset += 1;
-            update_batch_size(self.sequence_ranges, self.offset, &mut self.batch_size);
-        }
-        Some(index)
     }
 
     fn size_hint(&self) -> (usize, Option<usize>) {
-        let size = self
-            .sequence_ranges
-            .iter()
-            .take(self.batch_size)
-            .map(|range| range.len() - self.offset)
-            .sum::<usize>()
-            - self.seq_idx;
+        // Total size of all elements at index `offset` or later.
+        let level_size: usize = self
+            .sequences
+            .clone()
+            .map(|seq| seq.len().saturating_sub(self.offset))
+            .take_while(|&size| size > 0)
+            .sum();
+        let size = if level_size == 0 {
+            // This is handled because sequences_iter ends up incorrectly indicating that one
+            // element has been emitted when the iterator has been fully exhausted.
+            0
+        } else {
+            // Subtract the number of elements emitted so far at this offset level.
+            level_size - (self.sequences.len() - self.sequences_iter.len())
+        };
         (size, Some(size))
     }
 }
 
-impl<'a> ExactSizeIterator for PackingIndices<'a> {}
-
-/// The batch size of each offset when sequences of the given lengths are packed.
-///
-/// # Example
-/// ```
-/// use relearn::utils::packed::PackedBatchSizes;
-///
-/// let sequence_lengths = [4, 2, 2];
-/// let batch_sizes: Vec<_> = PackedBatchSizes::from_sorted_lengths(&sequence_lengths).collect();
-/// assert_eq!(batch_sizes, vec![3, 3, 1, 1]);
-/// ```
-#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
-pub struct PackedBatchSizes<'a, T> {
-    /// Sequences lengths sorted in monotonically decreasing order.
-    sequence_lengths: &'a [T],
-    /// Current offset within the sequences.
-    offset: usize,
-    /// Number of sequences with length > offset
-    batch_size: usize,
-}
-
-impl<'a> PackedBatchSizes<'a, usize> {
-    /// Create a [`PackedBatchSizes`] instance from monotonically decreasing sequence lengths.
-    pub fn from_sorted_lengths(sequence_lengths: &'a [usize]) -> Self {
-        let offset = 0;
-        let mut batch_size = sequence_lengths.len();
-        update_batch_size(sequence_lengths, offset, &mut batch_size);
-        Self {
-            sequence_lengths,
-            offset,
-            batch_size,
-        }
-    }
-}
-
-impl<'a> PackedBatchSizes<'a, Range<usize>> {
-    /// Create a [`PackedBatchSizes`] instance from ranges with monotonically decreasing lengths.
-    pub fn from_sorted_ranges(sequence_ranges: &'a [Range<usize>]) -> Self {
-        let offset = 0;
-        let mut batch_size = sequence_ranges.len();
-        update_batch_size(sequence_ranges, offset, &mut batch_size);
-        Self {
-            sequence_lengths: sequence_ranges,
-            offset,
-            batch_size,
-        }
-    }
-}
-
-impl<'a, T> Iterator for PackedBatchSizes<'a, T>
+impl<I> ExactSizeIterator for PackedSeqIter<I>
 where
-    T: Length,
+    I: Iterator + ExactSizeIterator + Clone,
+    <I as Iterator>::Item: Deref,
+    <<I as Iterator>::Item as Deref>::Target: Sequence,
 {
-    type Item = usize;
-
-    fn next(&mut self) -> Option<Self::Item> {
-        if self.batch_size == 0 {
-            None
-        } else {
-            let current_batch_size = self.batch_size;
-            self.offset += 1;
-            update_batch_size(self.sequence_lengths, self.offset, &mut self.batch_size);
-            Some(current_batch_size)
-        }
-    }
-
-    fn size_hint(&self) -> (usize, Option<usize>) {
-        let length = match self.sequence_lengths.first() {
-            Some(v) => v.length() - self.offset,
-            None => 0,
-        };
-        (length, Some(length))
-    }
-}
-
-impl<'a, T> ExactSizeIterator for PackedBatchSizes<'a, T> where T: Length {}
-
-/// Decrement batch size until it is the batch size of the given offset.
-fn update_batch_size<T>(sorted_sequence_ranges: &T, offset: usize, batch_size: &mut usize)
-where
-    T: Index<usize> + ?Sized,
-    <T as Index<usize>>::Output: Length,
-{
-    while *batch_size > 0 && sorted_sequence_ranges[*batch_size - 1].length() <= offset {
-        *batch_size -= 1;
-    }
-}
-
-/// Has a length
-pub trait Length {
-    fn length(&self) -> usize;
-}
-
-impl Length for Range<usize> {
-    fn length(&self) -> usize {
-        self.len()
-    }
-}
-
-/// Directly stores a length
-impl Length for usize {
-    fn length(&self) -> usize {
-        *self
-    }
 }
 
 /// View each batched offset of a packed tensor.
@@ -511,12 +440,11 @@ pub fn packed_tensor_discounted_cumsum_from_end(
 #[cfg(test)]
 mod packed_iter {
     use super::*;
-    use std::array::IntoIter;
 
     #[test]
     fn iter_from_sorted() {
         let sequences = [vec![0, 1, 2, 3], vec![10, 11], vec![100, 101]];
-        let packed: Vec<_> = PackedIter::from_sorted(IntoIter::new(sequences)).collect();
+        let packed: Vec<_> = PackedIter::from_sorted(sequences).collect();
         let expected = vec![0, 10, 100, 1, 11, 101, 2, 3];
         assert_eq!(packed, expected);
     }
@@ -524,7 +452,7 @@ mod packed_iter {
     #[test]
     fn iter_from_unsorted() {
         let sequences = [vec![10, 11], vec![0, 1, 2, 3], vec![100, 101]];
-        let packed: Vec<_> = PackedIter::new(IntoIter::new(sequences)).collect();
+        let packed: Vec<_> = PackedIter::new(sequences).collect();
         // First sorts the sequences then packs them.
         let expected = vec![0, 10, 100, 1, 11, 101, 2, 3];
         assert_eq!(packed, expected);
@@ -556,7 +484,7 @@ mod packing_indices {
     fn iter() {
         let data = [0, 1, 2, 3, 10, 11, 100, 101];
         let ranges = [0..4, 4..6, 6..8];
-        let packed: Vec<_> = PackingIndices::from_sorted(&ranges)
+        let packed: Vec<_> = PackedSeqIter::from_sorted(&ranges)
             .map(|i| data[i])
             .collect();
         let expected = vec![0, 10, 100, 1, 11, 101, 2, 3];
@@ -566,14 +494,14 @@ mod packing_indices {
     #[test]
     fn size_hint() {
         let ranges = [0..4, 4..6, 6..8];
-        let packing_indices = PackingIndices::from_sorted(&ranges);
+        let packing_indices = PackedSeqIter::from_sorted(&ranges);
         assert_eq!(packing_indices.size_hint(), (8, Some(8)));
     }
 
     #[test]
     fn size_hint_after_next() {
         let ranges = [0..4, 4..6, 6..8];
-        let mut packing_indices = PackingIndices::from_sorted(&ranges);
+        let mut packing_indices = PackedSeqIter::from_sorted(&ranges);
         let _ = packing_indices.next();
         assert_eq!(packing_indices.size_hint(), (7, Some(7)));
         let _ = packing_indices.next();
@@ -581,44 +509,6 @@ mod packing_indices {
     }
 }
 
-#[cfg(test)]
-mod packed_batch_sizes {
-    use super::*;
-
-    #[test]
-    fn iter_from_lengths() {
-        let sequence_lengths = [4, 2, 2];
-        let packed: Vec<_> = PackedBatchSizes::from_sorted_lengths(&sequence_lengths).collect();
-        let expected = vec![3, 3, 1, 1];
-        assert_eq!(packed, expected);
-    }
-
-    #[test]
-    fn iter_from_ranges() {
-        let ranges = [0..4, 4..6, 6..8];
-        let packed: Vec<_> = PackedBatchSizes::from_sorted_ranges(&ranges).collect();
-        let expected = vec![3, 3, 1, 1];
-        assert_eq!(packed, expected);
-    }
-
-    #[test]
-    fn size_hint() {
-        let sequence_lengths = [4, 2, 2];
-        let packed_batch_sizes = PackedBatchSizes::from_sorted_lengths(&sequence_lengths);
-        assert_eq!(packed_batch_sizes.size_hint(), (4, Some(4)));
-    }
-
-    #[test]
-    fn size_hint_after_next() {
-        let sequence_lengths = [4, 2, 2];
-        let mut packed_batch_sizes = PackedBatchSizes::from_sorted_lengths(&sequence_lengths);
-        let _ = packed_batch_sizes.next();
-        assert_eq!(packed_batch_sizes.size_hint(), (3, Some(3)));
-        let _ = packed_batch_sizes.next();
-        assert_eq!(packed_batch_sizes.size_hint(), (2, Some(2)));
-    }
-}
-
 #[cfg(test)]
 mod packed_tensor {
     use super::*;
